/*
*  Name:      abstract_connection_manager.hh
*  Author:    Rafael Jesus Alcantara Perez
*  Summary:   Class for handling connections in C++.
*  Date:      $Date: 2003/10/06 13:21:35 $
*  Revision:  $Revision: 1.1 $
*
*  Copyright (C) 1999-2003  Rafael Jesus Alcantara Perez <rafa@dedalo-ing.com>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
*  MA 02111-1307, USA.
*/

#ifndef _UESQL_ABSTRACT_CONNECTION_MANAGER__
#define _UESQL_ABSTRACT_CONNECTION_MANAGER__

#include <algorithm>
#include <mpcl/exceptions.hh>
#include <mpcl/text/string.hh>
#include <mpcl/util/collection/map.hh>
#ifdef _UESQL_DEBUG__
#include <mpcl/util/logging/log.hh>
#endif  // not _UESQL_DEBUG__
#include <queue>
#include <vector>
#include "defs.hh"


///  Name-space for UESQLC.
namespace uesqlc
{

  ///  Unary function object to delete parameters.
  template <typename TItem>
  struct MUpdateFromBuffer : public std::unary_function<TItem, void>
  {

      void operator () (TItem* ptITEM) const
      {
        ptITEM->updateFromBuffer();
      }

  };  // struct MUpdateFromBuffer

  ///  Unary function object to delete parameters.
  template <typename TItem>
  struct MWriteDebugInfo : public std::unary_function<TItem, void>
  {

      void operator () (TItem* ptITEM) const
      {
#       ifdef _UESQL_DEBUG__
        mpcl::util::logging::tLog << ", " << ptITEM->get();
#       endif  // not _UESQL_DEBUG__
      }

  };  // struct MWriteDebugInfo

  /**
  *  UESQL base manager class. The methods required for type \a TConnectionData are:
  *
  *    TConnectionData::TConnectionData (const TConnectionData& rktCONNECTION);
  *    void TConnectionData::close();
  */
  template <typename TConnectionData, typename TResult>
  class TAbstractConnectionManager : protected TSqlstateDescriptor
  {

    protected:

      ///  Map type that binds connection names to connections.
      typedef
        mpcl::util::collection::TMap<mpcl::text::TString, TConnectionData>
        TConnectionDataMap;

      ///  Integer with the ordinal of the next unique connection name.
      mutable int   iUniqueOrdinal;

      ///  Map that binds connection names to connections.
      TConnectionDataMap   tConnectionDataMap;

      ///  Current connection (from \a tConnectionDataMap).
      TConnectionData*   ptCurrentConnection;

      ///  Input embedded variable string values queue.
      std::queue<mpcl::text::TString>   tInputEmbeddedVariableQueue;

      ///  Output embedded variable string pointers vector.
      std::vector<IType*>   tOutputEmbeddedVariableVector;

      ///  String with the identifier of the current connection.
      mpcl::text::TString   yCurrentConnection;


    public:

      //
      //  C O N S T R U C T O R S
      //

      ///  Builds a new instance.
      TAbstractConnectionManager (void)      :
        TSqlstateDescriptor           ()     ,
        iUniqueOrdinal                (0)    ,
        tConnectionDataMap            ()     ,
        ptCurrentConnection           (NULL) ,
        tInputEmbeddedVariableQueue   ()     ,
        tOutputEmbeddedVariableVector ()     ,
        yCurrentConnection            ()
      {
        initialize();
      }

      ///  Destroys the instance.
      virtual ~TAbstractConnectionManager (void)
      {
        finish();
      }

      /**
      *  Binds in connection map, connection name \a rkyCONNECTION_NAME to
      *  \a rtCONNECTION.
      *  @param rkyCONNECTION_NAME Connection name.
      *  @param rtCONNECTION       Connection.
      */
      void bindConnection ( const mpcl::text::TString& rkyCONNECTION_NAME ,
                            TConnectionData&           rtCONNECTION       )
      {
        tConnectionDataMap.bind (rkyCONNECTION_NAME, rtCONNECTION);
      }

      /**
      *  Diagnoses statement  result  (\a tRESULT). It only cheks the result of
      *  the statement executed.  This member function (\a diagnose())  updates
      *  properly SQL status variables.
      *  @param tRESULT Source statement execution result.
      */
      virtual void diagnose (TResult tRESULT) = 0;

      /**
      *  Unbinds connection named \a rkyCONNECTION_NAME.
      *  @param rkyCONNECTION_NAME Source connection name.
      *  @warning kyCONNECTION_NAME  mustn't  be  passed  as reference to avoid
      *           problems  about passing  \a yCurrentConnection  as  argument.
      */
      void unbindConnection (const mpcl::text::TString kyCONNECTION_NAME)
      {

        bool                                    gIsCurrentConnection;
        typename TConnectionDataMap::iterator   I = tConnectionDataMap.find (kyCONNECTION_NAME);

        if ( I == tConnectionDataMap.end() )
        {
          throw mpcl::TConstraintException ("there is no such connection", __FILE__, __LINE__);
        }
        else
        {
          gIsCurrentConnection = ( ptCurrentConnection == &(I->second) );
          I->second.close();
          tConnectionDataMap.erase (I);
          if ( gIsCurrentConnection )
          {
            yCurrentConnection.erase();
            ptCurrentConnection = NULL;
          }
        }

      }  // unbindConnection()

      ///  Unbinds current connection (if any).
      void unbindCurrentConnection (void)
      {
        if ( existsDefaultConnection() )
        {
          unbindConnection (yCurrentConnection);
        }
      }

      ///  Unbinds all connections.
      void unbindAllConnections (void)
      {
        while ( !tConnectionDataMap.empty() )
        {
          unbindConnection (tConnectionDataMap.begin()->first);
        }
      }

      /**
      *  Clears  embedded  variable  containers.  It  must  be  called  at  the
      *  beginning of each new instruction.
      */
      virtual void clearQueues (void)
      {
        while ( !tInputEmbeddedVariableQueue.empty() )
        {
          tInputEmbeddedVariableQueue.pop();
        }
        tOutputEmbeddedVariableVector.clear();
      }

      /**
      *  Clears the result of the last executed statement. The
      *  diagnostic  member  functions,  use  that results  to
      *  diagnose the execution.
      */
      virtual void clearResult (TResult tRESULT) = 0;

      /**
      *  Executes the statement rkySTATEMENT instantiating
      *  the input embedded variables (tagged with '$z').
      *  @param rkySTATEMENT String with the source statement.
      */
      virtual TResult execute (const mpcl::text::TString& rkySTATEMENT)
      {
        return execute (rkySTATEMENT.c_str());
      }

      /**
      *  Executes the statement  \a rkySTATEMENT instantiating the input
      *  embedded variables (tagged with '$z') in the \a determinate connection
      *  named \a rkyCONNECTION_NAME.
      *  @param rkySTATEMENT       String with the source statement.
      *  @param rkyCONNECTION_NAME Source connection name.
      */
      virtual TResult execute ( const mpcl::text::TString& rkySTATEMENT       ,
                                const mpcl::text::TString& rkyCONNECTION_NAME )
      {
        return execute (rkySTATEMENT.c_str(), rkyCONNECTION_NAME.c_str());
      }

      /**
      *  Executes the statement pkcSTATEMENT instantiating
      *  the input embedded variables (tagged with '$z').
      *  @param pkcSTATEMENT String with the source statement.
      */
      virtual TResult execute (const char* pkcSTATEMENT) = 0;

      /**
      *  Executes the statement  \a pkcSTATEMENT instantiating the input
      *  embedded variables (tagged with '$z') in the \a determinate connection
      *  named \a rkyCONNECTION_NAME.
      *  @param pkcSTATEMENT       String with the source statement.
      *  @param pkcCONNECTION_NAME Source connection name.
      */
      virtual TResult execute ( const char* pkcSTATEMENT       ,
                                const char* pkyCONNECTION_NAME ) = 0;

      /**
      *  Executes the statement pkcSTATEMENT instantiating
      *  the input embedded variables (tagged with '$z'). The result
      *  of the execution is omitted.
      *  @param pkcSTATEMENT String with the source statement.
      */
      virtual void executeClean (const char* pkcSTATEMENT)
      {
        clearResult (execute (pkcSTATEMENT));
      }

      ///  Finishes all managed issues.
      virtual void finish (void)
      {
        unbindAllConnections();
        clearQueues();
        tConnectionDataMap.clear();
        yCurrentConnection.erase();
      }

      /**
      *  Gets a pointer to the SQL data type of the output embedded variable at
      *  position \a zINDEX in the output embedded variable vector.  It must be
      *  called for writing results on output embedded variables.
      *  @param zINDEX Position of the variable.
      *  @return Pointer to the output embedded variable.
      */
      virtual IType* getFromOutputQueue (const size_t zINDEX) const
      {
        if ( zINDEX >= tOutputEmbeddedVariableVector.size() )
        {
          throw mpcl::TConstraintException ("out of range", __FILE__, __LINE__);
        }
        return tOutputEmbeddedVariableVector [zINDEX];
      }

      ///  Initializes the default connection.
      virtual void initialize (void) {}

      /**
      *  Returns a string that is result of instanting the input
      *  embedded variables (if any).
      *  @param pkcSTATEMENT Source statement.
      *  @return The instantiated string.
      */
      mpcl::text::TString instantiate (const char* pkcSTATEMENT);

      /**
      *  Returns a string that is result of instanting the input
      *  embedded variables (if any).
      *  @param rkySTATEMENT Source statement.
      *  @return The instantiated version of string \a rkySTATEMENT.
      */
      mpcl::text::TString instantiate (const mpcl::text::TString& rkySTATEMENT)
      {
        return instantiate (rkySTATEMENT.c_str());
      }

      /**
      *  Pushes an input embedded variable value into the input queue.
      *  @param rktEMBEDDED_VALUE Input embedded variable value.
      */
      virtual void pushOnInputQueue (const IType& rktEMBEDDED_VALUE)
      {
        tInputEmbeddedVariableQueue.push (rktEMBEDDED_VALUE.get());
      }

      /**
      *  Pops from the output embedded variable vector the last variable.
      *  @return The last added embedded variable reference.
      */
      virtual IType* popFromOutputQueue (void)
      {
        IType*   ptBaseType = tOutputEmbeddedVariableVector.back();

        tOutputEmbeddedVariableVector.pop_back();
        return ptBaseType;
      }

      /**
      *  Pushes on the output embedded variable vector the variable
      *  reference \a rtSQL_DATA_TYPE.
      *  @param rtSQL_DATA_TYPE Output embedded variable reference.
      */
      virtual void pushOnOutputQueue (IType* ptSQL_DATA_TYPE)
      {
        tOutputEmbeddedVariableVector.push_back (ptSQL_DATA_TYPE);
      }

      ///  Resets SQL status variables.
      void resetStatus (void)
      {
        //
        //  Condition:  successful completion
        //
        SQLCODE    = 0;
        SQLMESSAGE = "";
        SQLSTATE   = "00000";
      }

      /**
      *  Sets current connection to connection named rkyCONNECTION_NAME.
      *  @param rkyCONNECTION_NAME Source connection name.
      */
      void setCurrentConnection (const mpcl::text::TString& rkyCONNECTION_NAME);

      /**
      *  Updates output variables  (which  have  been updated passively using a
      *  temporary buffer).
      */
      virtual void updateOutputVariables (void);


    public:

      //
      //  S E L E C T O R S
      //

      /**
      *  Returns the current connection data.
      *  @return The current connection data.
      */
      TConnectionData& getCurrentConnectionData (void)
      {
        if ( ptCurrentConnection == NULL )
        {
          throw mpcl::TConstraintException ("there is no current connection", __FILE__, __LINE__);
        }
        return *ptCurrentConnection;
      }

      /**
      *  Returns the connection name for the current connection.
      *  @return The connection name for the current connection.
      */
      const mpcl::text::TString& currentConnectionName (void) const
      {
        return yCurrentConnection;
      }

      /**
      *  Returns true if there is a default connection.
      *  @return True if there is a default connection.
      */
      virtual bool existsDefaultConnection (void) const
      {
        return ( ptCurrentConnection != NULL );
      }

      /**
      *  Returns true if there is an active connection.
      *  @param rkyCONNECTION_NAME Source connection name to check.
      *  @return True if there is an active connection.
      */
      bool isConnected (const mpcl::text::TString& rkyCONNECTION_NAME) const
      {
        return tConnectionDataMap.isBound (rkyCONNECTION_NAME);
      }

      /**
      *  Checks if a NOT FOUND condition has happened.
      *  @return True if it has happened.
      */
      bool isNotFoundCondition (void) const
      {
        return ( SQLSTATE == no_data::pkcNO_SUBCLASS );
      }

      /**
      *  Checks if a SQLERROR condition has happened.
      *  @return True if it has happened.
      */
      bool isSqlerrorCondition (void) const
      {
        return ( SQLCODE < 0 );
      }

  };  // class TAbstractConnectionManager

}  // namespace uesqlc


//
//  C O N S T R U C T O R S
//

template <typename TConnectionData, typename TResult>
inline mpcl::text::TString uesqlc::TAbstractConnectionManager<TConnectionData, TResult>::
instantiate (const char* pkcSTATEMENT)
{

  using mpcl::text::TString;

  register TString::size_type   zPatternOffset;
  TString                       yStatementInstance (pkcSTATEMENT);

  while ( !tInputEmbeddedVariableQueue.empty() )
  {
    if ( TString::npos == (zPatternOffset = yStatementInstance.find ("$z")) )
    {
      break;
    }
    else
    {
      yStatementInstance.replace (zPatternOffset, 2, tInputEmbeddedVariableQueue.front());
      tInputEmbeddedVariableQueue.pop();
    }
  }
  return yStatementInstance;

}  // instantiate()


template <typename TConnectionData, typename TResult>
inline void uesqlc::TAbstractConnectionManager<TConnectionData, TResult>::
setCurrentConnection (const mpcl::text::TString& rkyCONNECTION_NAME)
{

  if ( tConnectionDataMap.isBound (rkyCONNECTION_NAME) )
  {
    yCurrentConnection  = rkyCONNECTION_NAME;
    ptCurrentConnection = &(tConnectionDataMap [rkyCONNECTION_NAME]);
#   ifdef _UESQL_DEBUG__
    mpcl::util::logging::tLog << mpcl::util::logging::header << "SET CONNECTION '" << rkyCONNECTION_NAME << "';\n";
#   endif  // not _UESQL_DEBUG__
    resetStatus();
  }
  else
  {
    //
    //  Condition:           connection exception
    //  Class subcondition:  connection does not exist
    //
    SQLCODE    = -1;
    SQLMESSAGE = "";
    SQLSTATE   = "08003";
  }

}  // setCurrentConnection()


template <typename TConnectionData, typename TResult>
inline void uesqlc::TAbstractConnectionManager<TConnectionData, TResult>::
updateOutputVariables (void)
{

# ifdef _UESQL_DEBUG__
  using mpcl::util::logging::header;
  using mpcl::util::logging::tLog;
# endif  // not _UESQL_DEBUG__
  using std::for_each;

  for_each ( tOutputEmbeddedVariableVector.begin() ,
             tOutputEmbeddedVariableVector.end()   ,
             MUpdateFromBuffer<IType>()            );
# ifdef _UESQL_DEBUG__
  tLog << header << "FETCHED (";
  if ( !tOutputEmbeddedVariableVector.empty() )
  {
    tLog << tOutputEmbeddedVariableVector.front()->get();
    for_each ( tOutputEmbeddedVariableVector.begin() + 1 ,
               tOutputEmbeddedVariableVector.end()       ,
               MWriteDebugInfo<IType>()                  );
  }
  tLog << ");\n";
# endif  // not _UESQL_DEBUG__

}  // updateOutputVariables()


#endif  // not _UESQL_ABSTRACT_CONNECTION_MANAGER__
